// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __ABSTRACT_MRF_CU__
#define __ABSTRACT_MRF_CU__

#include <iostream>

#include <rfdk/GPUGraph.hpp>
#include <rfdk/GPUParameter.hpp>

// the macro EXXN5(n) will expand to an if-statement with 32 branches, where each branch performs another if-statement with 32 branches. Thus 1024 branches in total.

#define EXN0(A,B) else if(deg == A && X == B){cudaFuncSetCacheConfig(computeMessagesLoopy<R, B, A, doMAP>, cudaFuncCachePreferL1);computeMessagesLoopy<R, B, A, doMAP><<<NLBP, threads>>>(*G, *P, num, i, G->getNodeWithNeighborCountArray(i), in_msg, out_msg, obs);}
#define EXN1(n,m) EXN0(n,m) EXN0(n+1,m)
#define EXN2(n,m) EXN1(n,m) EXN1(n+2,m)	
#define EXN3(n,m) EXN2(n,m) EXN2(n+4,m)
#define EXN4(n,m) EXN3(n,m) EXN3(n+8,m)
#define EXN5(n,m) EXN4(n,m) EXN4(n+16,m)
#define EXXN0(n) EXN4(1,n)
#define EXXN1(n) EXXN0(n) EXXN0(n+1)
#define EXXN2(n) EXXN1(n) EXXN1(n+2)
#define EXXN3(n) EXXN2(n) EXXN2(n+4)
#define EXXN4(n) EXXN3(n) EXXN3(n+8)
#define EXXN5(n) EXXN4(n) EXXN4(n+16)

#define EXXP0(n) else if(X == n){cudaFuncSetCacheConfig(postMessages<R,n>, cudaFuncCachePreferL1);postMessages<R,n><<<blocksPM, threads>>>(*G, in_msg, out_msg);}
#define EXXP1(n) EXXP0(n) EXXP0(n+1)
#define EXXP2(n) EXXP1(n) EXXP1(n+2)
#define EXXP3(n) EXXP2(n) EXXP2(n+4)
#define EXXP4(n) EXXP3(n) EXXP3(n+8)
#define EXXP5(n) EXXP4(n) EXXP4(n+16)

#define EXX0(n) else if(X == n){cudaFuncSetCacheConfig(computeMessagesLoopy<R,n,doMAP>, cudaFuncCachePreferL1);computeMessagesLoopy<R,n,doMAP><<<blocksLBP, threads>>>(*G, *P, in_msg, out_msg, obs);}
#define EXX1(n) EXX0(n) EXX0(n+1)
#define EXX2(n) EXX1(n) EXX1(n+2)
#define EXX3(n) EXX2(n) EXX2(n+4)
#define EXX4(n) EXX3(n) EXX3(n+8)
#define EXX5(n) EXX4(n) EXX4(n+16)

namespace MRF{

	extern unsigned int VERBOSEBP;

	template <class R>
	__device__ R accumulateProd(R a, R b) {
		return a+b;
	}

	template <class R>
	__device__ R accumulateMax(R a, R b) {
		return max(a,b);
	}

	template <class R>
	__device__ R accumulateSum(R a, R b) {
		return a+b;
	}

	template <class R>
	__device__ R potential(R a, R b) {
		return exp(a+b);
	}

	template <class R>
	__device__ R result(R a) {
		return std::log(a);
	}

	template <class T>
	__device__  T getLogPotential(GPUParameterStorage<T> P, const unsigned int i) {
		return P.get(i);
	}

	template <class T>
	__device__  T getLogPotential(GPUParameterStorage<T> P, const unsigned int s, const unsigned int x, const unsigned int X) {
		return getLogPotential<T>(P, s*X+x);
	}

	template <class T>
	__device__  T getLogPotential(GPUParameterStorage<T> P, const unsigned int eid, const unsigned int x, const unsigned int y, const unsigned int X, const unsigned int n) {
		return getLogPotential<T>(P, n*X+eid*X*X+x*X+y);
	}

#ifdef __UNROLL_XD__
	template <class T, unsigned int X, unsigned int N, bool doMAP>
	__device__ T msg(const unsigned int eid, const unsigned int x, bool reverse, const GPUGraph G, GPUParameterStorage<T>& P, T* in_msg, int* O){
#elif __UNROLL_X__
	template <class T, unsigned int X, bool doMAP>
	__device__ T msg(const unsigned int eid, const unsigned int x, bool reverse, const GPUGraph G, GPUParameterStorage<T>& P, T* in_msg, int* O){
#else
	template <class T, bool doMAP>
	__device__ T msg(const unsigned int eid, const unsigned int x, const unsigned int X, bool reverse, const GPUGraph G, GPUParameterStorage<T>& P, T* in_msg, int* O){
#endif
		unsigned int s, t;

		const unsigned int n = G.getNumNodes();
		const unsigned int m = G.getNumEdges();

		if(!reverse){
			s = G.getEdgeSource(eid);
			t = G.getEdgeTarget(eid);
		}else{
			s = G.getEdgeTarget(eid);
			t = G.getEdgeSource(eid);
		}

		const int o = O[s]; // load observed value (is -1 if node is unobserved)

		T _prod, sum = 0;
		if(o == -1){
#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
#pragma unroll
#endif
			for(unsigned int k=0; k<X; ++k){
				_prod = 0;
#ifdef __UNROLL_XD__
#pragma unroll
#else
				const unsigned int N = G.getNumNeighbors(s);
#endif
				for(unsigned int i = 0; i < N; ++i){
					if(G.getNeighbor(s,i) != t){
						T in;
						const unsigned int e = G.getNeighborEdge(s,i);
						if(G.getEdgeSource(e) == s) in = in_msg[m*X + e*X+k];
						else in = in_msg[e*X+k];

						_prod = accumulateProd(_prod, in);
					}
				}
				T theta_s = getLogPotential<T>(P, s, k, X), theta_st;

				if(reverse)	theta_st = getLogPotential<T>(P, eid, x, k, X, n);
				else		theta_st = getLogPotential<T>(P, eid, k, x, X, n);

				if(doMAP)	sum = accumulateMax(sum, potential(theta_s+theta_st, _prod)); // Max-Product
				else		sum = accumulateSum(sum, potential(theta_s+theta_st, _prod)); // Sum-Product
			}
		}else{
			T theta_st;

			if(reverse)	theta_st = getLogPotential<T>(P, eid, x, o, X, n);
			else		theta_st = getLogPotential<T>(P, eid, o, x, X, n);

			sum = potential(getLogPotential<T>(P, s, o, X)+theta_st, T(0));
		}
		return result(sum);
	}


#ifdef __UNROLL_XD__
	template <class R, unsigned int X, unsigned int D, bool doMAP>
	__global__ void computeMessagesLoopy(GPUGraph G, GPUParameterStorage<R> P, unsigned int n, unsigned int j, unsigned int* jump, R* in_msg, R* out_msg, int* O){
#elif __UNROLL_X__
	template <class R, unsigned int X, bool doMAP>
	__global__ void computeMessagesLoopy(GPUGraph G, GPUParameterStorage<R> P, R* in_msg, R* out_msg, int* O){
#else
	template <class R, bool doMAP>
	__global__ void computeMessagesLoopy(GPUGraph G, GPUParameterStorage<R> P, const unsigned int X, R* in_msg, R* out_msg, int* O){
#endif
		const unsigned int id = blockIdx.x * blockDim.x + threadIdx.x; // id = i*X + x (i<n*D) (id<n*D*X)
		const unsigned int m = G.getNumEdges();
#ifdef __UNROLL_XD__
		if(id < n*D*X){
#else
		if(id < 2*m*X){
#endif
			// decode thread index
			const unsigned int x   = id%X;
#ifdef __UNROLL_XD__
			const unsigned int i   = (id-x)/X;
			unsigned int       eid = jump[i];
			const unsigned int r   = (eid>=m);
			if(r) eid -= m;
#else
			const unsigned int eid	= ((id-x)%(m*X))/X;
			const unsigned int r	= (id-(eid*X+x))/(m*X);
#endif

			// compute all messages in parallel
#ifdef __UNROLL_XD__
			out_msg[r*m*X + eid*X + x] = msg<R, X, D, doMAP>(eid, x, r, G, P, in_msg, O);
#elif __UNROLL_X__
			out_msg[r*m*X + eid*X + x] = msg<R, X, doMAP>(eid, x, r, G, P, in_msg, O);
#else
			out_msg[r*m*X + eid*X + x] = msg<R, doMAP>(eid, x, X, r, G, P, in_msg, O);
#endif
		}
	}

	template <class R>
	__global__ void clearMessages(GPUGraph G, R* in_msg, R* out_msg, unsigned int X, unsigned int dim){

		const unsigned int id = blockIdx.x * blockDim.x + threadIdx.x; // id = r*m*X + eid*X + x

		if(id < dim){
			// decode thread index
			const unsigned int x = id%X;
			const unsigned int m = G.getNumEdges();
			const unsigned int eid = ((id-x)/X)%m;
			const unsigned int r = (id-(x+eid*X))/(m*X);

			in_msg[r*m*X + eid*X + x]  = 0;
			out_msg[r*m*X + eid*X + x] = 0;
		}
	}

#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
	template <class R, unsigned int X>
	__device__ void postMessage(unsigned int i, bool reverse, unsigned int m, R* in_msg, R* out_msg, R alpha){
#else
	template <class R>
	__device__ void postMessage(unsigned int i, bool reverse, unsigned int m, unsigned int X, R* in_msg, R* out_msg, R alpha){
#endif
		R _z = 0;
		unsigned int o = 0;
		if(reverse) o = m*X;
#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
#pragma unroll
#endif
		for(unsigned int k=0; k<X; ++k){
			_z += exp(out_msg[o+i*X+k]);
		}
#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
#pragma unroll
#endif
		for(unsigned int k=0; k<X; ++k){
			const unsigned int idx = o+i*X+k;
			R t = in_msg[idx];
			in_msg[idx] = log(alpha*exp(t) + (1-alpha)*exp(out_msg[idx] - log(_z)));
//			in_msg[idx] = alpha*t + (1-alpha)*(out_msg[idx] - log(_z));
			out_msg[idx] = t;
		}
	}

#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
	template <class R, unsigned int X>
	__global__ void postMessages(GPUGraph G, R* in_msg, R* out_msg, R alpha){
#else
	template <class R>
	__global__ void postMessages(GPUGraph G, unsigned int X, R* in_msg, R* out_msg, R alpha){
#endif

		const unsigned int id = blockIdx.x * blockDim.x + threadIdx.x; // id = i*2+r
		const unsigned int m = G.getNumEdges();

		if(id < m*2){
			// decode thread index
			const unsigned int r = id%2;
			const unsigned int i = (id-r)/2;
#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
			postMessage<R, X>(i, r, m, in_msg, out_msg, alpha);
#else
			postMessage<R>(i, r, m, X, in_msg, out_msg, alpha);
#endif
		}
	}

	template <class R>
	__device__ void smemReduce(volatile R* shared, unsigned int tid){
		if (tid < 32){
			shared[tid] += shared[tid + 32];
			shared[tid] += shared[tid + 16];
			shared[tid] += shared[tid +  8];
			shared[tid] += shared[tid +  4];
			shared[tid] += shared[tid +  2];
			shared[tid] += shared[tid +  1];
		}
	}

	template <class R>
	__global__ void l2diff(R* d, R* a, R* b, unsigned int n){

		__shared__ R shared[128];

		unsigned int tid      = threadIdx.x;
		unsigned int i        = blockIdx.x*256 + threadIdx.x;
		unsigned int gridSize = 256*gridDim.x;

		shared[tid] = 0.0f;

		unsigned int idx = 0;

		while(i < n){
			R val = a[i] - b[i];
	     		shared[tid] += val*val;

	 		if(i + 128 < n){
				idx = i + 128;
				R val = a[idx] - b[idx];
				shared[tid] += val*val;
			}
			i += gridSize;
		}
		__syncthreads();

		if (tid <  64) shared[tid] += shared[tid +  64];
		__syncthreads();

		smemReduce(shared, tid);
		if (tid == 0) d[blockIdx.x] = shared[0];
	}

	template <class R>
	__global__ void reduceL2diff(R* data){

		__shared__ R shared[64];

		const unsigned int tid = threadIdx.x;
		const unsigned int i = threadIdx.x;
		shared[tid] = data[i] + data[i+64];
		__syncthreads();
		smemReduce(shared, tid);

		if (tid == 0) data[tid] = sqrt(shared[tid]);
	}

	template <class R, bool doMAP>
	void computeMessagesLoopy_cuda(AbstractGraph*& _G, AbstractParameterStorage<R>*& _P, R* in, R* out,  R*& in_msg, R*& out_msg, R*& result, const unsigned int X, const unsigned int J, int* O, int*& obs, R alpha){

		//if(MRF::VERBOSEBP) std::cout << "computeMessagesLoopy_cuda(" << doMAP << ")" << std::endl;

		const unsigned int n = _G->getNumNodes();
		const unsigned int m = _G->getNumEdges();

		R eps	= 0.01;
		R d	= 1;

		GPUGraph* G = dynamic_cast<GPUGraph*>(_G);
		if(!G){
			 G = new GPUGraph(_G);
			_G = G;
		}

		GPUParameterStorage<R>* P = new GPUParameterStorage<R>(_P);

//		GPUParameterStorage<R>* P = new GPUParameterStorage<R>(_P);

//		R* temp = new R[100];
//		cudaMemcpy(temp, P->getArray(), 100*sizeof(R), cudaMemcpyDeviceToHost);
//		for(unsigned int i=0; i<100; ++i){
//			std::cout << i << '\t' << _P->getArray()[i] << '\t' << temp[i] << std::endl;
//		}
//		delete[] temp;


		const unsigned int s = max(2*m*X, 256);
		if(!in_msg)  cudaMalloc((void**)&in_msg,  s*sizeof(R));
		if(!out_msg) cudaMalloc((void**)&out_msg, s*sizeof(R));
		if(!obs) cudaMalloc((void**)&obs, n*sizeof(R));
		if(!result)  cudaHostAlloc((void**)&result, 128*sizeof(R), cudaHostAllocMapped);

		cudaMemcpy(obs, O, n*sizeof(int), cudaMemcpyHostToDevice);

		const unsigned int threads = 192;
		unsigned int blocksCLR = (s+threads-1)/threads;
#ifndef __UNROLL_XD__
		unsigned int blocksLBP = (2*m*X+threads-1)/threads;
#endif
		unsigned int blocksPM  = (m*2+threads-1)/threads;

		cudaFuncSetCacheConfig(clearMessages<R>, cudaFuncCachePreferL1);
		clearMessages<<<blocksCLR, threads>>>(*G, in_msg, out_msg, X, s);

		for(unsigned int j=0; j<J && d>eps; ++j){

			for(unsigned int i=0; i<128; ++i) result[i] = 0;
#ifdef __UNROLL_XD__
			for(int i=0; i<G->getNumNeighborCounts(); ++i){
				const unsigned int deg = G->getNumNeighborSize(i);
				const unsigned int num = G->getNumNeighborCounts(i);
				const unsigned int NLBP = (deg*num*X+threads-1)/threads;
				if(X>8 || deg>16) break;
				EXXN3(2);
			}
#elif __UNROLL_X__
			if(X>32) break;
			EXX5(2);
#else
			cudaFuncSetCacheConfig(computeMessagesLoopy<R, doMAP>, cudaFuncCachePreferL1);
			computeMessagesLoopy<R, doMAP><<<blocksLBP, threads>>>(*G, *P, X, in_msg, out_msg, obs);
#endif

#if defined(__UNROLL_XD__) || defined(__UNROLL_X__)
			if(X>32) break;
			EXXP5(2);
#else
			cudaFuncSetCacheConfig(postMessages<R>, cudaFuncCachePreferL1);
			postMessages<R><<<blocksPM, threads>>>(*G, X, in_msg, out_msg, alpha);
#endif
			l2diff<R><<<128, 128>>>(result, in_msg, out_msg, 2*m*X);

			cudaDeviceSynchronize();

			d = 0;
			for(unsigned int i=0; i<128; ++i) d += result[i];
			d = sqrt(d);

			if(MRF::VERBOSEBP) std::cout << j << "\tdiff = " << d << std::endl;
		}

		cudaMemcpy(in,  in_msg,  2*m*X*sizeof(R), cudaMemcpyDeviceToHost);
		cudaMemcpy(out, out_msg, 2*m*X*sizeof(R), cudaMemcpyDeviceToHost);

		P->finish();
		delete P;
	}

	template <class R>
	void free_cuda(R*& in, R*& out, R*& r, int*& o, AbstractGraph*& _G, AbstractParameterStorage<R>*& _P){
		cudaFree(in);
		cudaFree(out);
		cudaFree(o);
		cudaFreeHost(r);

		GPUGraph* GG = dynamic_cast<GPUGraph*>(_G);
		if(GG){
			((GPUGraph*)GG)->finish();
			delete ((GPUGraph*)GG);
		}
		GPUParameterStorage<R>* PP = dynamic_cast<GPUParameterStorage<R>*>(_P);
		if(PP){
			((GPUParameterStorage<R>*)PP)->finish();
			delete ((GPUParameterStorage<R>*)PP);
		}
	}

	template <>
	void free_cuda<float>(float*& in, float*& out, float*& r, int*& o, AbstractGraph*& _G, AbstractParameterStorage<float>*& _P){
		cudaFree(in);
		cudaFree(out);
		cudaFree(o);
		cudaFreeHost(r);

		GPUGraph* GG = dynamic_cast<GPUGraph*>(_G);
		if(GG){
			((GPUGraph*)GG)->finish();
			delete ((GPUGraph*)GG);
		}
		GPUParameterStorage<float>* PP = dynamic_cast<GPUParameterStorage<float>*>(_P);
		if(PP){
			((GPUParameterStorage<float>*)PP)->finish();
			delete ((GPUParameterStorage<float>*)PP);
		}
	}

	template <>
	void free_cuda<double>(double*& in, double*& out, double*& r, int*& o, AbstractGraph*& _G, AbstractParameterStorage<double>*& _P){
		cudaFree(in);
		cudaFree(out);
		cudaFree(o);
		cudaFreeHost(r);

		GPUGraph* GG = dynamic_cast<GPUGraph*>(_G);
		if(GG){
			((GPUGraph*)GG)->finish();
			delete ((GPUGraph*)GG);
		}
		GPUParameterStorage<double>* PP = dynamic_cast<GPUParameterStorage<double>*>(_P);
		if(PP){
			((GPUParameterStorage<double>*)PP)->finish();
			delete ((GPUParameterStorage<double>*)PP);
		}
	}

	// template instantiation - do not try to call this method!
	void __DO_NOT_CALL__(){

		float  *F = 0;
		double *D = 0;
		int    *I = 0;

		AbstractGraph* G = 0;
		AbstractParameterStorage<float>*  Pf = 0;
		AbstractParameterStorage<double>* Pd = 0;
/*
		free_cuda<float>(F, F, F, I, G, Pf);
		free_cuda<double>(D, D, D, I, G, Pd);
*/
		computeMessagesLoopy_cuda<float, false>(G, Pf, F, F, F, F, F, 0, 0, I, I, *F);
		computeMessagesLoopy_cuda<float, true >(G, Pf, F, F, F, F, F, 0, 0, I, I, *F);
		computeMessagesLoopy_cuda<double,false>(G, Pd, D, D, D, D, D, 0, 0, I, I, *D);
		computeMessagesLoopy_cuda<double,true >(G, Pd, D, D, D, D, D, 0, 0, I, I, *D);
	}
}
#endif
