#include <iostream>
#include <iomanip>
#include <fstream>

#include <rfdk/01Graph.hpp>

#include <strf/STRF.hpp>
#include <strf/STCSVReader.hpp>
 
int main(int argc, char** argv){

	std::cout << "STRF learner 06/03/2014, (c) 2013,2014, Nico Piatkowski, all rights reserved." << std::endl;

	::MRF::VERBOSE = false;
	::MRF::VERBOSEBP = false;

	if(argc%2==0){
		std::cout << "Usage: " << argv[0] << " --opt1 val1 --opt2 val2 .. --optN valN" << std::endl;
		return 5;
	}

	double lam1 = 0;
	double lam2 = 0;
	double dimin = 0.00000001;
	double dimax = 1;
	double dinc = 0.0001;
	double alpha = 0;
	double damp = 0.5;
	double ceps = 0.000001;
	double vertexMinFreq = 0.0;

	unsigned int T = 24;
	unsigned int loopy = 200;

	size_t maxIter = -1;

	bool useVertexWeights = true;
	bool useSpatialEdgeWeights = true;

	std::string fNameG = "";
	std::string fNameMin = "";
	std::string fNameMout = "";
	std::string fNameOG = "";
	std::string fNameV = "";
	std::string fNameEMP = "";
	std::string fNameEST = "";
	std::string fD = "";
	std::string sTimeKey = "t";
	std::string sPrimaryKey = "id";
	std::string sStates = "0,1";

	for(unsigned int i=1; i<argc; i+=2){
		std::string a(argv[i]);
		if(!a.compare("--graph"))	fNameG = std::string(argv[i+1]);
		else if(!a.compare("--data"))	fD = std::string(argv[i+1]);
		else if(!a.compare("--vertexset"))	fNameV = std::string(argv[i+1]);
		else if(!a.compare("--model"))	fNameMin = std::string(argv[i+1]);
		else if(!a.compare("--omodel"))	fNameMout = std::string(argv[i+1]);
		else if(!a.compare("--ograph"))	fNameOG = std::string(argv[i+1]);
		else if(!a.compare("--timekey"))	sTimeKey = std::string(argv[i+1]);
		else if(!a.compare("--primary"))	sPrimaryKey = std::string(argv[i+1]);
		else if(!a.compare("--states"))	sStates = std::string(argv[i+1]);
		else if(!a.compare("--emp"))	fNameEMP = std::string(argv[i+1]);
		else if(!a.compare("--est"))	fNameEST = std::string(argv[i+1]);
		else if(!a.compare("--alpha"))	alpha = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--lam1"))	lam1 = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--lam2"))	lam2 = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--dimin"))	dimin = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--dimax"))	dimax = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--goal"))	dinc = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--damp"))	damp = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--minfreq"))	vertexMinFreq = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--convergence"))	ceps = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--maxiter"))	maxIter = MRF::get<size_t>(argv[i+1]);
		else if(!a.compare("--T"))	T = MRF::get<size_t>(argv[i+1]);
		else if(!a.compare("--lbpiter"))	loopy = MRF::get<size_t>(argv[i+1]);
		else if(!a.compare("--v0"))	MRF::VERBOSE = MRF::get<unsigned int>(argv[i+1]);
		else if(!a.compare("--v1"))	MRF::VERBOSEBP = MRF::get<unsigned int>(argv[i+1]);
		else if(!a.compare("--spatialw"))	useSpatialEdgeWeights = MRF::get<unsigned int>(argv[i+1]);
		else if(!a.compare("--vertexw"))	useVertexWeights = MRF::get<unsigned int>(argv[i+1]);

		else{
//			showUsage(argv);
			return -1;
		}
	}

	if(!fD.compare("")){
		std::cout << "Please provide training data with: " << argv[0] << " --data <.csv-file>" << std::endl;
		return -1;
	}

	MRF::Graph *H = 0;

	if(fNameG.compare("")){
		std::cout << "Loading graph from " << fNameG << std::endl;
		std::ifstream Gin(fNameG);
		H = new MRF::Graph(Gin);
		Gin.close();

	}else if(fNameV.compare("")){
		std::cout << "Building 0-1 graph (alpha = " << alpha << ", V = " << fNameV << ", vMinFreq =  " << vertexMinFreq << ")";
		///////////////////////////////////////////////////////////////////////////////
		// 0-1 graph
		///////////////////////////////////////////////////////////////////////////////

		std::vector<std::string> vertexNames;
		std::ifstream nv(fNameV);

		while(!nv.eof()){
			std::string line;
			getline(nv,line);
			if(line.size()>1) vertexNames.push_back(line);
		}
		nv.close();

		H = new MRF::ZeroOneGraph(vertexNames, fD, alpha, vertexMinFreq);

		if(fNameOG.compare("")){
			std::cout << "Storing 0-1 graph at " << fNameOG << std::endl;
			std::ofstream Gout(fNameOG);
			Gout << *H;
			Gout.close();
		}
	}else{
		std::cout << "Please provide either a graph file with: " << argv[0] << 
		" --graph <.graph-file>" << std::endl << "or a vertex set file with: " << argv[0] << " --vertexset <.V-file>" << std::endl;
		return -1;
	}

	MRF::STGraph* G = new MRF::STGraph(H, T);
	MRF::STCSVReader* reader = new MRF::STCSVReader(G);

	reader->setTimeKey(sTimeKey);	// the name of the column that contains the integer time index

	// READ STATES FROM COMMAND LINE
	size_t pos = sStates.find(','),lpos = 0;
	while(pos != std::string::npos){
		std::string state = sStates.substr(lpos,pos-lpos);
		reader->addState(state);
		lpos = pos+1;
		pos = sStates.find(',',lpos);
	}
	reader->addState(sStates.substr(lpos));

	// READ PK FROM COMMAND LINE
	pos = sPrimaryKey.find(','),lpos = 0;
	while(pos != std::string::npos){
		std::string state = sPrimaryKey.substr(lpos,pos-lpos);
		reader->addPrimary(state);
		lpos = pos+1;
		pos = sPrimaryKey.find(',',lpos);
	}
	reader->addPrimary(sPrimaryKey.substr(lpos));

	// READ TRAINING DATA AND COMPUTE STATISTICS
	reader->loadData(fD);

	const unsigned int dim = reader->getDim();

	const unsigned int N = reader->getNumInstances();
	const unsigned int X = reader->getNumStates();

	double* F = reader->getProbs();

	if(fNameEMP.compare("")){
		std::cout << "Storing empirical vertex marginals at " << fNameEMP << std::endl;

		for(unsigned int j=0; j<X; ++j){
			std::stringstream fsin;
			fsin << fNameEMP << '.';
			fsin << reader->getStateName(j);

			std::ofstream eO(fsin.str());
			for(unsigned int i=0; i<G->getSpatialNodes(); ++i){
				eO << G->getVertexNameById(i);
				for(unsigned int t=0; t<T; ++t){
					eO << ',' << F[i*T*X+t*X+j];
				}
				eO << std::endl;
			}
			eO.close();
		}
	}

	std::cout << "Configuration:" << std::endl;
	std::cout << "   - T " << T << std::endl;
	std::cout << "   - dimension " << dim << std::endl;
	std::cout << "   - lambda_1 " << std::fixed << std::setprecision(8) << lam1 << std::endl;
	std::cout << "   - lambda_2 " << lam2 << std::endl;
	std::cout << "   - max LBP iterations " << loopy << std::endl;
	std::cout << "   - LBP message damping " << damp << std::endl;
	std::cout << "   - min D^-1 " << dimin << std::endl;
	std::cout << "   - max D^-1 " << dimax << std::endl;
	std::cout << "   - convergence at " << ceps << std::endl;
	std::cout << "   - linesearch goal " << dinc << std::endl;
	std::cout << "   - max iterations " << ((long)maxIter) << std::endl;
	std::cout << "   - spatial edge weights " << useSpatialEdgeWeights << std::endl;
	std::cout << "   - vertex weights " << useVertexWeights << std::endl;
	std::cout << std::endl;

	MRF::STRF<double> learner((MRF::STGraph*)G, X, T);
	learner.zeroWeights();
	learner.setLoopyIterations(loopy);
	learner.useCompressedParameters(true);
	learner.setLambda1(lam1);
	learner.setLambda2(lam2);
	learner.setDinvmin(dimin);
	learner.setDinvmax(dimax);
	learner.setDesiredIncrement(dinc);
	learner.setEps(ceps);
	learner.setMaxIter(maxIter);
	learner.setDamping(damp);
	learner.useSpatialEdgeWeights(useSpatialEdgeWeights);
	learner.useVertexWeights(useVertexWeights);

	if(fNameMin.compare("")){
		std::cout << "Loading model parameters from " << fNameMin << std::endl;
		std::ifstream pIn(fNameMin);
		learner.loadParameters(pIn);
		pIn.close();
	}

	std::cout << "Learning.." << std::endl;
	double _obj = learner.optimize(F, 1);

	if(fNameEST.compare("")){
		std::cout << "Storing empirical vertex marginals at " << fNameEST << std::endl;

		F = learner.getProbs();

		for(unsigned int j=0; j<X; ++j){
			std::stringstream fsn;
			fsn << fNameEST << '.';
			fsn << reader->getStateName(j);

			std::ofstream eO(fsn.str());
			for(unsigned int i=0; i<G->getSpatialNodes(); ++i){
				eO << G->getVertexNameById(i);
				for(unsigned int t=0; t<T; ++t){
					eO << ',' << F[i*T*X+t*X+j];
				}
				eO << std::endl;
			}
			eO.close();
		}
	}

	if(fNameMout.compare("")){
		std::cout << "Storing model parameters at " << fNameMout << std::endl;
		std::ofstream pO(fNameMout);
		learner.storeParameters(pO);
		pO.close();
	}

	delete reader;
	delete G;

	return 0;
}
