#include <iostream>
#include <iomanip>
#include <fstream>

#include <rfdk/01Graph.hpp>

#include <strf/STRF.hpp>
#include <strf/STCSVReader.hpp>

int main(int argc, char** argv){

	std::cout << "STRF predictor 06/03/2014, (c) 2013,2014, Nico Piatkowski, all rights reserved." << std::endl;

	::MRF::VERBOSE = false;
	::MRF::VERBOSEBP = false;

	if(argc%2==0){
		std::cout << "Usage: " << argv[0] << " --opt1 val1 --opt2 val2 .. --optN valN" << std::endl;
		return 5;
	}

	double alpha = 0;
	double damp = 0.5;

	unsigned int T = 24;
	unsigned int begin = 0;
	unsigned int end = 24;
	unsigned int loopy = 200;

	std::string fNameG = "";
	std::string fNameMin = "";
	std::string fNameV = "";
	std::string fD = "";
	std::string sTimeKey = "t";
	std::string sPrimaryKey = "id";
	std::string sStates = "0,1";

	bool predictNext = false;

	for(unsigned int i=1; i<argc; i+=2){
		std::string a(argv[i]);
		if(!a.compare("--graph"))	fNameG = std::string(argv[i+1]);
		else if(!a.compare("--data"))	fD = std::string(argv[i+1]);
		else if(!a.compare("--vertexset"))	fNameV = std::string(argv[i+1]);
		else if(!a.compare("--model"))	fNameMin = std::string(argv[i+1]);
		else if(!a.compare("--timekey"))	sTimeKey = std::string(argv[i+1]);
		else if(!a.compare("--primary"))	sPrimaryKey = std::string(argv[i+1]);
		else if(!a.compare("--states"))	sStates = std::string(argv[i+1]);
		else if(!a.compare("--alpha"))	alpha = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--damp"))	damp = MRF::get<double>(argv[i+1]);
		else if(!a.compare("--T"))	T = MRF::get<size_t>(argv[i+1]);
		else if(!a.compare("--lbpiter"))	loopy = MRF::get<size_t>(argv[i+1]);
		else if(!a.compare("--v0"))	MRF::VERBOSE = MRF::get<unsigned int>(argv[i+1]);
		else if(!a.compare("--v1"))	MRF::VERBOSEBP = MRF::get<unsigned int>(argv[i+1]);
		else if(!a.compare("--pnext"))	predictNext = MRF::get<unsigned int>(argv[i+1]);

		else{
//			showUsage(argv);
			return -1;
		}
	}
	if(end>T) end = T;

	if(!fD.compare("")){
		std::cout << "Please provide test data with: " << argv[0] << " --data <.csv-file>" << std::endl;
		return -1;
	}

	MRF::Graph *H = 0;

	if(fNameG.compare("")){
		std::cout << "Loading graph from " << fNameG << std::endl;
		std::ifstream Gin(fNameG);
		H = new MRF::Graph(Gin);
		Gin.close();

	}else if(fNameV.compare("")){
		std::cout << "Building 0-1 graph (alpha = " << alpha << ", V = " << fNameV << ")";
		///////////////////////////////////////////////////////////////////////////////
		// 0-1 graph
		///////////////////////////////////////////////////////////////////////////////

		std::vector<std::string> vertexNames;
		std::ifstream nv(fNameV);

		while(!nv.eof()){
			std::string line;
			getline(nv,line);
			if(line.size()>1) vertexNames.push_back(line);
		}
		nv.close();

		H = new MRF::ZeroOneGraph(vertexNames, fD, alpha);
	}else{
		std::cout << "Please provide either a graph file with: " << argv[0] << 
		" --graph <.graph-file>" << std::endl << "or a vertex set file with: " << argv[0] << " --vertexset <.V-file>" << std::endl;
		return -1;
	}

	MRF::STGraph* G = new MRF::STGraph(H, T);
	MRF::STCSVReader* reader = new MRF::STCSVReader(G);

	reader->setTimeKey(sTimeKey);	// the name of the column that contains the integer time index

	// READ STATES FROM COMMAND LINE
	size_t pos = sStates.find(','),lpos = 0;
	while(pos != std::string::npos){
		std::string state = sStates.substr(lpos,pos-lpos);
		reader->addState(state);
		lpos = pos+1;
		pos = sStates.find(',',lpos);
	}
	reader->addState(sStates.substr(lpos));

	// READ PK FROM COMMAND LINE
	pos = sPrimaryKey.find(','),lpos = 0;
	while(pos != std::string::npos){
		std::string state = sPrimaryKey.substr(lpos,pos-lpos);
		reader->addPrimary(state);
		lpos = pos+1;
		pos = sPrimaryKey.find(',',lpos);
	}
	reader->addPrimary(sPrimaryKey.substr(lpos));

	// READ TEST DATA AND COMPUTE STATISTICS
	reader->loadData(fD);

	const unsigned int dim = reader->getDim();
	const unsigned int N = reader->getNumInstances();
	const unsigned int X = reader->getNumStates();

	double* F = reader->getProbs();

	std::cout << "Configuration:" << std::endl;
	std::cout << "   - T " << T << std::endl;
	std::cout << "   - dimension " << dim << std::endl;
	std::cout << "   - max LBP iterations " << loopy << std::endl;
	std::cout << "   - LBP message damping " << damp << std::endl;
	std::cout << std::endl;

	MRF::STRF<double> learner((MRF::STGraph*)G, X, T);
	learner.zeroWeights();
	learner.setLoopyIterations(loopy);
	learner.useCompressedParameters(true);
	learner.setDamping(damp);

	if(fNameMin.compare("")){
		std::cout << "Loading model parameters from " << fNameMin << std::endl;
		std::ifstream pIn(fNameMin);
		learner.loadParameters(pIn);
		pIn.close();
	}else{
		std::cout << "Please provide a trained model with: " << argv[0] << " --model <.mod-file>" << std::endl;
		return -1;
	}

	std::cout << "Predicting.." << std::endl;

	unsigned int C1[X][X];
	unsigned int C2[X][X];
	unsigned int RC[X], CC[X], OC[X];


	learner.MAP();
	unsigned int* mm = learner.getMAP();
	unsigned int mmap[G->getNumNodes()];
	for(unsigned int i=0; i<G->getNumNodes(); ++i) mmap[i] = mm[i];

	for(unsigned int f=begin; f<end-1; ++f){
		unsigned int hour = (f*60)/60;
		unsigned int min  = (f*60)%60;

		std::cout << std::setfill('0') << std::setw(2) << hour << ':' << std::setfill('0') << std::setw(2) << min << '\t';

		for(unsigned int i=0; i<X; ++i){
			RC[i] = 0;
			CC[i] = 0;
			OC[i] = 0;
			for(unsigned int j=0; j<X; ++j){
				C1[i][j] = 0;
				C2[i][j] = 0;
			}
		}


//		unsigned int f = begin;
//		if(true){
		for(unsigned int i=0; i<N; ++i){
			unsigned int hasData = 0;

			learner.clearObservations();

			for(unsigned int v=0; v<G->getSpatialNodes(); ++v) if(reader->getInstanceVertexState(i,v,f)!=0) hasData++;
			if(hasData){

//				std::cout << "INSTANCE " << reader->getInstanceName(i) << "(" << i << "); OBSERVED LAYER " << std::setfill('0') << std::setw(2) << hour << ':' << std::setfill('0') << std::setw(2) << min << "(" << f << ")" << std::endl;


				for(unsigned int v=0; v<G->getSpatialNodes(); ++v){
					learner.setObservation(v,f,reader->getInstanceVertexState(i,v,f));
				}

				learner.MAP();

				double tbat = 0.0;

				unsigned int target = end;
				if(predictNext) target = f+2;
				if(target>end) target = end;

				for(unsigned int t=f+1; t<target; ++t){
					unsigned int hour = (t*30)/60;
					unsigned int min  = (t*30)%60;

					unsigned int otherHasData = 0;
					for(unsigned int v=0; v<G->getSpatialNodes(); ++v) if(reader->getInstanceVertexState(i,v,t)!=0) otherHasData++;
					if(otherHasData){

						for(unsigned int v=0; v<G->getSpatialNodes(); ++v){
							int data = reader->getInstanceVertexState(i,v,t);
							int pred = learner.getMAPState(v,t);
							int opred = mmap[v*T+t];

							RC[data]++;
							CC[pred]++;
							OC[opred]++;
							C1[data][pred]++;
							C2[data][opred]++;
						}

/*
	//					std::cout << i << ' ' << t << "|" << f << ' ' << std::setfill('0') << std::setw(2) << hour << ':' << std::setfill('0') << std::setw(2) << min << " { ";

						std::stringstream ssb;
						ssb << "t_";
						ssb << t;

						double batdrop = rweights[ssb.str()] + rweights["_BIAS_"];

						for(unsigned int v=0; v<G->getSpatialNodes(); ++v){

							int data = reader->getInstanceVertexState(i,v,t);
							int pred = learner.getMAPState(v,t);

							if(pred){
								//std::cout << G->getVertexNameById(v) << '\t' << rweights[G->getVertexNameById(v)] << '\t' << batdrop << std::endl;
								double val = rweights[G->getVertexNameById(v)];
								batdrop += val;
							}

						}

						if(batdrop>0) batdrop = 0.0;
						tbat += batdrop;
						std::cout << "{ " << batdrop << " " << tbat << " }" << std::endl;
*/
					}
				}
			}
		}
		double diag1 = 0, other1 = 0, diag2 = 0, other2 = 0;
//		std::cout << std::endl;
//		std::cout << ' ' << '\t';
//		for(unsigned int i=0; i<X; ++i){
//			std::cout << i << '\t';
//		}
//		std::cout << "REC";
//		std::cout << std::endl;
		for(unsigned int i=0; i<X; ++i){
//			std::cout << i << '\t';
			for(unsigned int j=0; j<X; ++j){
//				std::cout << C[i][j] << '\t';
				if(i==j){
					diag1 += C1[i][j];
					diag2 += C2[i][j];
				}else{
					other1 += C1[i][j];
					other2 += C2[i][j];
				}
			}
//			std::cout << C[i][i]/(1.0*RC[i]) << '\t';
//			std::cout << std::endl;
		}
//		std::cout << "PREC" << '\t';
//		for(unsigned int i=0; i<X; ++i){
//			std::cout << C[i][i]/(1.0*CC[i]) << '\t';
//		}
		std::cout << double(diag1)/double(diag1+other1) << '\t' << double(C1[1][1])/double(RC[1]) << '\t' << double(diag2)/double(diag2+other2) << '\t' << double(C2[1][1])/double(RC[1]) << std::endl;
	}

	delete reader;
	delete G;

	return 0;
}
