// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __ABSTRACTPARAMETER_HPP__
#define __ABSTRACTPARAMETER_HPP__

namespace MRF{
	template <class T> class AbstractParameterStorage{

		protected:
			size_t n;

		public:
			AbstractParameterStorage(const size_t _n){
				n = _n;
			}

#ifdef __CUDACC__
			virtual __host__ __device__ T get(size_t i) const = 0;
			virtual __host__ __device__ void zero() = 0;
			virtual __host__ __device__ void set(const size_t i, const T val) = 0;
			virtual __host__ __device__ unsigned int getSize() const {return n;}
			virtual __host__ __device__  T* getArray(){return 0;};
#else
			virtual T get(size_t i) const = 0;
			virtual void zero() = 0;
			virtual void set(const size_t i, const T val) = 0;
			virtual unsigned int getSize() const {return n;}
			virtual T* getArray(){return 0;};
#endif
			virtual void finish(){};
	};
};
#endif
