// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __GPUPARAMETER_HPP__
#define __GPUPARAMETER_HPP__

#include <stdio.h>
#include <rfdk/AbstractParameter.hpp>

namespace MRF{
	template <class T> class GPUParameterStorage : public AbstractParameterStorage<T>{

		protected:
			using AbstractParameterStorage<T>::n;
			T* data;

		public:

			GPUParameterStorage(AbstractParameterStorage<T>* other) : AbstractParameterStorage<T>(other->getSize()){
				if(other){
					//T* temp = new T[n];
					//printf("DIM = %lu\n",n);
					cudaMalloc((void**)&data, n*sizeof(T));

					//for(unsigned int i=0; i<n; ++i){
					//	temp[i] = other->get(i);
					//}
					cudaMemcpy(data, other->getArray(), n*sizeof(T), cudaMemcpyHostToDevice);
					cudaDeviceSynchronize();
					//cudaMemcpy(data, temp, n*sizeof(T), cudaMemcpyHostToDevice);
					//delete temp;
				}else{
					n = 0;
				}
			}

			GPUParameterStorage(GPUParameterStorage<T>* other) : AbstractParameterStorage<T>(other->getSize()){
				data = other->getArray();
			}

			void finish(){
				cudaFree(data);
			}

			virtual __host__ __device__ void zero(){
				for(unsigned int i=0; i<n; ++i) data[i] = 0;
			}

			virtual __host__ __device__ T get(size_t i) const {
				return data[i];
			}

			virtual __host__ __device__ void set(const size_t i, const T val) {
				data[i] = val;
			}

			virtual __host__ __device__  T* getArray(){
				return data;
			}
	};
};
#endif
