// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __PARAMETER_HPP__
#define __PARAMETER_HPP__

#include <cstring>

#include <rfdk/AbstractParameter.hpp>

namespace MRF{
	template <class T> class ParameterStorage : public AbstractParameterStorage<T>{

		protected:
			using AbstractParameterStorage<T>::n;
			T* data;

		public:

			ParameterStorage(const size_t _n) : AbstractParameterStorage<T>(_n){
				//std::cout << "ParameterStorage: n = " << n << std::endl;
				data = new T[n];
				zero();
			}
/*
			ParameterStorage() : AbstractParameterStorage<T>(0){
				data = 0;
			}
*/
			~ParameterStorage(){
				//std::cout << "DELETING ParameterStorage::data at " << data << std::endl;
				if(data) delete[] data;
			}

			virtual T* getArray(){
				return data;
			}

			virtual void zero(){
				memset(data, 0, n*sizeof(T));
			}

			virtual T get(size_t i) const {
				return data[i];
			}

			virtual void set(const size_t i, const T val) {
				data[i] = val;
			}

			virtual T& operator[](const size_t i){
				return data[i];
			}
	};
};
#endif
