// Copyright (c) 2013,2014, Nico Piatkowski
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//    This product includes software developed by the Artificial Intelligence
//    Group of the TU Dortmund University, Germany.
// 4. Neither the name of the Artificial Intelligence Group, TU Dortmund nor the
//    names of its contributors may be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY NICO PIATKOWSKI ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL NICO PIATKOWSKI BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef __ST_CSV_READER_HPP__
#define __ST_CSV_READER_HPP__

#include <map>
#include <unordered_map>
#include <vector>
#include <algorithm>
#include <sstream>
#include <fstream>
#include <cassert>

#include <strf/STGraph.hpp>

namespace MRF{
	class STCSVReader{
		private:
			STGraph* G; // the graph

			std::vector<std::string> fields;
			std::unordered_map<std::string, std::unordered_map<unsigned int, std::unordered_map<std::string, std::string> > > data;

			std::string timeKey;
			std::set<std::string> PK;  // primary key
			std::vector<std::string> pks; // primary keys found in the data

			std::vector<std::string> stateSpace;
			std::unordered_map<std::string, unsigned int> stateMap;

			unsigned int X;
			unsigned int T;
			unsigned int n;
			unsigned int m;
			unsigned int dim;


			unsigned int* counts;
			double* probs;
			std::unordered_map<std::string, std::unordered_map<unsigned int, unsigned int> > totals;

			// warning: this is a duplicate of "encode" from SparseSTParameterStorage
			unsigned int encode(const int x,const int y,const int i,const int j,const int t,const int k) const {

				const int XX = X*X;
				const int nTX = n*T*X;
				const int nTm1XX = n*(T-1)*XX;

				if(i!=j && k!=-1){
					return nTX + n*(T-1)*XX + i*((T-1)*3+1)*XX + t*3*XX + k*XX + x*X + y;
				}

				else if(j==-1 && k==-1){
					return i*T*X + t*X + x;
				}

				else if(i==j && k==-1){
					return nTX + i*(T-1)*XX + t*XX + x*X + y;
				}

				assert(false);

				return 0;
			}

			// warning: this is a duplicate of "decode" from SparseSTParameterStorage
			void decode(const unsigned int i, int& dx, int& dy, int& di, int& dj, int& dt, int& dk) const {

				int IDX;

				const int XX = X*X;
				const int nTX = n*T*X;
				const int nTm1XX = n*(T-1)*XX;

				assert(i < dim);

				if(i < nTX){ // i = j*T*X + t*X + x
					dx = i % X; // = x
					int temp = (i-dx)/X; // = j*T + t
					dt = temp % T; // = t
					di = (temp-dt)/T; // = j

					IDX = di*T*X + dt*X + dx;
				}

				else if(i < nTX + nTm1XX){ // i = n*T*X + j*(T-1)*(X*X) + t*(X*X) + x*X + y

					int temp = i-nTX; // = j*(T-1)*(X*X) + t*(X*X) + x*X + y
					dy = i % X; // = y

					dx = (temp-dy)%XX; // = x*X
					dx /= X; // = x

					dt = (temp-(dx*X+dy))/XX; // = j*(T-1) + t
					dt %= (T-1); // = t

					di = (((temp-(dx*X+dy))/XX)-dt)/(T-1); // = j
					dj = di;

					IDX = nTX + di*(T-1)*XX + dt*XX + dx*X + dy;
				}

				else{ // i = n*T*X + n*(T-1)*(X*X) + j*(3*(T-1)+1)*(X*X) + t*3*(X*X) + k*(X*X) + x*X + y

					int temp = i-(nTX+nTm1XX); // = j*(3*(T-1)+1)*(X*X) + t*3*(X*X) + k*(X*X) + x*X + y

					dy = i % X; // = y

					dx = (temp-dy)%XX; // = x*X
					dx /= X; // = x

					dt = (temp-(dx*X+dy))/XX; // = j*((T-1)*3+1) + t*3 + k
					dt %= ((T-1)*3+1); // if t==T-1 && k>0: dt = k-1, else: dt = t*3 + k
					dt /= 3; // = t + (k/3) = t

					dk = (temp-(dx*X+dy))/XX; // = j*((T-1)*3+1) + t*3 + k
					dk %= ((T-1)*3+1); // = t*3 + k
					dk -= dt*3;

/*
					dk = (temp-(dx*X+dy))/XX; // = j*(3*(T-1)+1) + t*3 + k
					dk %= 3; // = k

					dt = (temp-(dx*X+dy+dk*XX))/XX; // = j*(3*(T-1)+1) + t*3
					dt %= (3*(T-1)+1); // = t*3
					dt /= 3;
*/

					di = (temp-(dx*X+dy+dk*XX+dt*3*XX))/(((T-1)*3+1)*XX); // = j
					
					IDX = nTX + nTm1XX + di*((T-1)*3+1)*XX + dt*3*XX + dk*XX + dx*X + dy;
				}
				assert(i==IDX);
			}

			std::string cc(std::string a, std::string b){
				std::stringstream ccs;
				ccs << a;
				ccs << '_';
				ccs << b;
				return ccs.str();
			}

		public:
			STCSVReader(STGraph* _G){
				G = _G;
				T = G->getNumLayers();
				n = G->getSpatialNodes();
				m = G->getSpatialEdges();
//				std::cout << "GRAPH HAS " << n << " NODES AND " << m << " EDGES" << std::endl;
				timeKey = "t";
			}

			~STCSVReader(){
				if(counts) delete[] counts;
				if(probs) delete[] probs;
			}

			void loadData(std::string& filename){
				std::ifstream inRAW(filename);
				std::string line, header;

				bool foundTimeKey = false;

				getline(inRAW, header); // assume first line is header

				if(VERBOSE) std::cout << "READ HEADER" << std::endl;
				size_t pos = header.find(','), lpos = 0;
				while(pos != std::string::npos){
					std::string name = header.substr(lpos, pos-lpos);
					fields.push_back(name);
					if(!timeKey.compare(name)) foundTimeKey = true;
//					std::cout << name << std::endl;
					lpos = pos+1;
					pos = header.find(',', lpos);
				}
				std::string name = header.substr(lpos);
				fields.push_back(name);

				if(!foundTimeKey){
					std::cout << "Error: time key " << timeKey << " not found in data file " << filename << std::endl;
				}
				assert(foundTimeKey);

				std::string lastPK = "";

				// read data
				if(VERBOSE) std::cout << "READ DATA" << std::endl;
				while(getline(inRAW, line)){

					// load current line into map
					std::unordered_map<std::string, std::string> current;
					size_t i = 0;
					pos = line.find(','), lpos = 0;
					while(pos != std::string::npos){
						std::string val = line.substr(lpos, pos-lpos);
						current[fields[i]] = val;
						assert(val.size());
						lpos = pos+1;
						pos = line.find(',', lpos);
						++i;
					}
					std::string val = line.substr(lpos, pos-lpos);
					current[fields[i]] = val;

					// get time index
					unsigned int t = str2uint(current[timeKey]);

					// get primary key
					std::string pk = filename;
					for(auto ii : PK){
						pk += "_"+current[ii];
						if(!current[ii].size()){
							std::cout << "Error: primary key column " << ii << " not found in data file " << filename << std::endl;
							bool hasPrimaryKey = false;
							assert(hasPrimaryKey);
						}
					}

					if(pk.compare(lastPK)!=0){
						//std::cout << "FOUND INSTANCE " << pk << std::endl;
						lastPK = pk;
						pks.push_back(pk);
					}

					data[pk][t] = current;
				}

				inRAW.close();

				std::cout << "Found " << pks.size() << " instances" << std::endl;
				// now, detect the number of states per node (X)
				for(auto id : pks){ // for all instances
					if(VERBOSE) std::cout << "\t" << id << std::endl;
					for(unsigned int i=0; i<G->getSpatialNodes(); ++i){
						std::string vid = G->getVertexNameById(i);
						//if(VERBOSE) std::cout << "\t\t" << vid << std::endl;
						if(data[id][0].find(vid) == data[id][0].end()){
							//std::cerr << "WARNING: THERE IS NO DATA FOR NODE " << vid << std::endl;
						}
						//if(data[id][0].find(ii.first) == data[id][0].end()){
						//	std::cerr << "WARNING: THERE IS NO DATA FOR NODE " << ii.first << std::endl;
						//}
						else 
						for(unsigned int t=0; t<T; ++t){
							if(data[id][t][vid].size()>0 && std::find(stateSpace.begin(), stateSpace.end(), data[id][t][vid]) == stateSpace.end()){
								stateSpace.push_back(data[id][t][vid]);
								if(VERBOSE) std::cout << "Found state " << data[id][t][vid] << " (t = " << t << ")" << std::endl;
							}
/*
							if(data[id][t][vid].size()<1){
								std::cout << "t = " << t << std::endl;
								assert(false);
							}
*/
						}
					}
				}

				X = stateSpace.size();

				std::cout << "Found " << X << " states: ";
				for(unsigned int x=0; x<X; ++x){
					stateMap[stateSpace[x]] = x;
					std::cout << " " << stateSpace[x] << "(" << stateMap[stateSpace[x]] << ")" << std::flush;
				}
				std::cout << std::endl;

				const unsigned int XX = X*X;
				const unsigned int nTX = n*T*X;
				const unsigned int nTm1XX = n*(T-1)*XX;

				dim = nTX+nTm1XX+m*(3*(T-1)+1)*XX;

				counts = new unsigned int[dim];
				probs = new double[dim];

				for(unsigned int idx = 0; idx < nTX; ++idx) counts[idx] = 0; // X;
				for(unsigned int idx = nTX; idx < dim; ++idx) counts[idx] = 0; //1;

				if(VERBOSE) std::cout << "Counting vertex states" << std::endl;
				// now, iterate over the spatio-temporal graph and compute empirical counts for every vertex and edge
				// count spatio-temporal node states and temporal edge states

				double start = getTime();
//#pragma omp parallel for
				for(unsigned int i=0; i<n; ++i){
					std::string vid = G->getVertexNameById(i);

					double pdone = 100.0*(i+1.0)/n; // precent done since start
					double tdone = getTime() - start; // time elapsed since start
					double ETA = (100.0/(pdone/tdone)) - tdone;

					char timeSymb = 's';
					if(ETA>180.0){
						ETA /= 60.0;
						timeSymb = 'm';
					}

					if(VERBOSE) std::cout << std::fixed << std::setprecision(3) << "\r\033[K" << pdone << "% [" << ETA << timeSymb << "]" << std::flush;

					std::string idx_temp_s = cc(vid,vid);

					for(unsigned int t = 0; t<T; ++t){

						for(auto id : pks){
							std::string xs = data[id][t][vid];
							unsigned int x = stateMap[xs];

							if(xs.size() > 0){
								unsigned int idx = encode(x,-1,i,-1,t,-1);
								++counts[idx];
								++totals[vid][t];
								//std::cout << vid << '\t' << t << '\t' << x << '\t' << counts[idx] << '\t' << totals[vid][t] << std::endl;
							}

							// temporal edge for node i
							if(t<T-1){
								std::string ys = data[id][t+1][vid];

								if(ys.size() > 0){
									unsigned int y = stateMap[ys];
									unsigned int idx = encode(x,y,i,i,t,-1);

									++counts[idx];
									++totals[idx_temp_s][t];
								}
							}
						}
					}
				}
				if(VERBOSE) std::cout << std::endl;
				if(VERBOSE) std::cout << "Counting edge states.." << std::endl;
				// count spatio-temporal edge states
//#pragma omp parallel for
				start = getTime();
				for(unsigned int i=0; i<m; ++i){
					unsigned int v = G->getEdgeSource(i)%T;
					unsigned int u = G->getEdgeTarget(i)%T;

					double pdone = 100.0*(i+1.0)/m; // precent done since start
					double tdone = getTime() - start; // time elapsed since start
					double ETA = (100.0/(pdone/tdone)) - tdone;

					char timeSymb = 's';
					if(ETA>180.0){
						ETA /= 60.0;
						timeSymb = 'm';
					}

					if(VERBOSE) std::cout << std::fixed << std::setprecision(3) << "\r\033[K" << pdone << "% [" << ETA << timeSymb << "]" << std::flush;

					std::string vid = G->getVertexNameById(v);
					std::string uid = G->getVertexNameById(u);

					std::string idx_st_s1 = cc(vid,uid);
					std::string idx_st_s2 = cc(cc(vid,uid),"1");
					std::string idx_st_s3 = cc(cc(vid,uid),"2");

					//std::cout << idx_st_s1 << "   " << idx_st_s2 << "   " << idx_st_s3 << std::endl;

					for(unsigned int t = 0; t<T; ++t){
						for(auto id : pks){

							std::string xs = data[id][t][vid];
							std::string ys = data[id][t][uid];

							if(xs.size() > 0 && ys.size() > 0){
								unsigned int x = stateMap[xs];
								unsigned int y = stateMap[ys];
								unsigned int idx = encode(x,y,i,-1,t,0);

								++counts[idx]; // spatial edge
								++totals[idx_st_s1][t];
							}

							if(t<T-1){

								//xs = data[id][t][vid];
								ys = data[id][t+1][uid];

								if(xs.size() > 0 && ys.size() > 0){
									unsigned int x = stateMap[xs];
									unsigned int y = stateMap[ys];
									unsigned int idx = encode(x,y,i,-1,t,1);

									++counts[idx]; // spatio-temporal edge 1
									++totals[idx_st_s2][t];
								}

								xs = data[id][t+1][vid];
								ys = data[id][t][uid];

								if(xs.size() > 0 && ys.size() > 0){
									unsigned int x = stateMap[xs];
									unsigned int y = stateMap[ys];
									unsigned int idx = encode(x,y,i,-1,t,2);

									++counts[idx]; // spatio-temporal edge 2
									++totals[idx_st_s3][t];
								}
							}
						}
					}
				}
				if(VERBOSE) std::cout << std::endl;
				if(VERBOSE) std::cout << "Counting done!" << std::endl;
			}

			std::string getStateName(unsigned int i){
				return stateSpace[i];
			}

			std::string getInstanceName(unsigned int i){
				return pks.at(i);
			}

			unsigned int numData(unsigned int i, unsigned int t){
				return data[pks.at(i)][t].size();
			}

			int getInstanceVertexState(unsigned int i, unsigned int v, unsigned int t){
				return stateMap[data[pks.at(i)][t][G->getVertexNameById(v)]];
			}

			// warning: the result has to be deleted by caller!
			unsigned int* getInstance(unsigned int i){
				unsigned int* result = new unsigned int[n*T];
				for(unsigned int j=0; j<n; ++j)
					for(unsigned int t=0; t<T; ++t)
						result[j*T+t] = getInstanceVertexState(i,j,t);
				return result;
			}

			unsigned int* getCounts(){
				return counts;
			}

			double* getProbs(){
/*
				std::cout << "=== DEBUG COUNTS IN PROBS ===" << std::endl;
				for(unsigned int i=0; i<n; ++i){
					std::string vid = G->getVertexNameById(i);
					for(unsigned int t = 0; t<T; ++t){
						for(unsigned int x=0; x<X; ++x){
							unsigned int idx = encode(x,-1,i,-1,t,-1);

						}
					}
				}
				std::cout << "=== END DEBUG COUNTS ===" << std::endl;
*/
 				if(VERBOSE) std::cout << "VERTEX PROBS" << std::endl;
//#pragma omp parallel for
				for(unsigned int i=0; i<n; ++i){
					std::string vid = G->getVertexNameById(i);
					for(unsigned int t = 0; t<T; ++t){
						unsigned int q = totals[vid][t];
						for(unsigned int x=0; x<X; ++x){
							unsigned int idx = encode(x,-1,i,-1,t,-1);
							
							//std::cout << '(' << i << ',' << t << ',' << x << ')' << '\t' << counts[idx] << '\t' << totals[vid][t] << std::endl;

							if(q!=0) probs[idx] = counts[idx]/(1.0*q);
							else probs[idx] = 1.0/X;
							//if(i==38){
							//std::cout << idx << '\t' << vid << '\t' << t << '\t' << x << '\t' << probs[idx] << '\t' << counts[idx] << '\t' << q << std::endl;
							//}

							// temporal edge for node i
							if(t<T-1){
								for(unsigned int y=0; y<X; ++y){
									unsigned int idx = encode(x,y,i,i,t,-1);

									if(totals[cc(vid,vid)][t]!=0) probs[idx] = counts[idx]/(1.0*totals[cc(vid,vid)][t]);
									else probs[idx] = 1.0/(X*X);
								}
							}
						}
					}
				}
 				if(VERBOSE) std::cout << "EDGE PROBS" << std::endl;
				// count spatio-temporal edge states
#pragma omp parallel for
				for(unsigned int i=0; i<m; ++i){
					unsigned int v = G->getEdgeSource(i)%T;
					unsigned int u = G->getEdgeTarget(i)%T;

					std::string vid = G->getVertexNameById(v);//nN[v];
					std::string uid = G->getVertexNameById(u);//nN[u];
					for(unsigned int t = 0; t<T; ++t){
						for(unsigned int x=0; x<X; ++x){
							for(unsigned int y=0; y<X; ++y){
								unsigned int idx = encode(x,y,i,-1,t,0);

								double z = totals[cc(vid,uid)][t];
								if(z!=0.0) probs[idx] = counts[idx]/z;
								else probs[idx] = 1.0/(X*X);

								if(t<T-1){
									idx = encode(x,y,i,-1,t,1);
									double z = totals[cc(cc(vid,uid),"1")][t];
									if(z!=0.0) probs[idx] = counts[idx]/z;
									else probs[idx] = 1.0/(X*X);

									idx = encode(x,y,i,-1,t,2);
									z = totals[cc(cc(vid,uid),"2")][t];
									if(z!=0.0) probs[idx] = counts[idx]/z;
									else probs[idx] = 1.0/(X*X);
								}
							}
						}
					}
				}
 				if(VERBOSE) std::cout << "PROBS DONE" << std::endl;
				return probs;
			}

			void addState(std::string s){
				stateSpace.push_back(s);
			}

			unsigned int getDim(){
				return dim;
			}

			unsigned int getNumStates(){
				return X;
			}

			unsigned int getNumInstances(){
				return pks.size();// + X*X;
			}

			void setTimeKey(std::string s){
				timeKey = s;
			}

			void addPrimary(std::string k){
				PK.insert(k);
			}
	};
};
#endif
